# Copyright 2020 Google Inc. All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import

from six.moves import urllib


def _add_set_cookie(request, value):
    request.extra_headers.append(('Set-Cookie', value))


def web_socket_do_extra_handshake(request):
    components = urllib.parse.urlparse(request.uri)
    command = components[4]

    ONE_DAY_LIFE = 'Max-Age=86400'

    if command == 'set':
        _add_set_cookie(request, '; '.join(['foo=bar', ONE_DAY_LIFE]))
    elif command == 'set_httponly':
        _add_set_cookie(
            request, '; '.join(['httpOnlyFoo=bar', ONE_DAY_LIFE, 'httpOnly']))
    elif command == 'clear':
        _add_set_cookie(request, 'foo=0; Max-Age=0')
        _add_set_cookie(request, 'httpOnlyFoo=0; Max-Age=0')


def web_socket_transfer_data(request):
    pass
